package eu.dnetlib.wfs.nodes;

import java.util.stream.Stream;

import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;

import eu.dnetlib.common.clients.DnetServiceClientFactory;
import eu.dnetlib.common.clients.MDStoreManagerClient;
import eu.dnetlib.common.clients.OaiManagerClient;
import eu.dnetlib.common.mdstores.backends.sql.MDStoreSqlBackend;
import eu.dnetlib.domain.dsm.Api;
import eu.dnetlib.domain.dsm.Datasource;
import eu.dnetlib.domain.mdstore.MDStoreVersion;
import eu.dnetlib.domain.mdstore.MDStoreWithInfo;
import eu.dnetlib.domain.mdstore.records.MetadataRecord;
import eu.dnetlib.domain.oai.OaiConfiguration;
import eu.dnetlib.domain.oai.OaiSet;
import eu.dnetlib.errors.DnetException;
import eu.dnetlib.wfs.annotations.WfInputParam;
import eu.dnetlib.wfs.annotations.WfNode;
import eu.dnetlib.wfs.output.oai.OaiBackend;
import jakarta.transaction.Transactional;

@WfNode("md_oai_export")
public class MdExportOaiJobNode extends ProcessNode {

	@WfInputParam
	private Datasource ds;

	@WfInputParam
	private Api api;

	@WfInputParam
	private String inputMdId;

	@WfInputParam
	private String oaiSetSpec;

	@WfInputParam
	private boolean oaiRefresh;

	@Autowired
	private DnetServiceClientFactory clientFactory;

	@Autowired
	private MDStoreSqlBackend mdStoreSqlBackend;

	@Autowired
	private OaiBackend oaiBackend;

	@Override
	@Transactional
	protected void execute() throws Exception {
		if (StringUtils.isBlank(oaiSetSpec)) {
			oaiSetSpec = DigestUtils.md5Hex(ds.getId()).substring(0, 8);
		}

		final OaiManagerClient oaiManager = clientFactory.getClient(OaiManagerClient.class);
		final MDStoreManagerClient mdstoreManager = clientFactory.getClient(MDStoreManagerClient.class);

		final OaiConfiguration conf = oaiManager.getOaiConfiguration();
		final MDStoreWithInfo mdstore = mdstoreManager.findMDStore(inputMdId);

		if (mdstore.getFormat().equals(conf.getNativeFormat())) {
			final boolean isSetRegistered = oaiManager.listSets()
					.stream()
					.map(OaiSet::getSetSpec)
					.filter(s -> s.equals(oaiSetSpec))
					.findFirst()
					.isPresent();

			if (!isSetRegistered) {
				final OaiSet oaiSet = new OaiSet();
				oaiSet.setSetSpec(oaiSetSpec);
				oaiSet.setSetName(ds.getOfficialname());
				oaiSet.setDescription("Publications of " + ds.getOfficialname());
				oaiSet.setDsId(ds.getId());
				oaiManager.createOaiSet(oaiSet);
			}

			final MDStoreVersion inputVersion = mdstoreManager.startReading(inputMdId);

			final Stream<MetadataRecord> records = mdStoreSqlBackend.streamEntries(inputVersion);
			try {
				oaiBackend.addRecords(records, oaiSetSpec, oaiRefresh);
			} finally {
				mdstoreManager.endReading(inputVersion);
			}
		}
		throw new DnetException("The native oai format (" + conf.getNativeFormat() + ") and the mdstore format (" + mdstore.getFormat() + ") are different");
	}

}
