package eu.dnetlib.wfs.collector.oai;

import java.io.StringReader;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Iterator;
import java.util.Queue;
import java.util.concurrent.PriorityBlockingQueue;

import org.apache.commons.lang3.StringUtils;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Node;
import org.dom4j.io.SAXReader;
import org.springframework.web.client.RestTemplate;

import eu.dnetlib.errors.CollectorException;
import eu.dnetlib.utils.XmlCleaner;

class OaiIterator implements Iterator<String> {

	private static final DateTimeFormatter oaiDateFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd");

	private final Queue<String> queue = new PriorityBlockingQueue<>();
	private final SAXReader reader = new SAXReader();

	private final String baseUrl;
	private final String set;
	private final String mdFormat;
	private final LocalDateTime fromDate;
	private final LocalDateTime untilDate;
	private String token;
	private boolean started;

	public OaiIterator(final String baseUrl,
			final String mdFormat,
			final String set,
			final LocalDateTime fromDate,
			final LocalDateTime untilDate) {
		this.baseUrl = baseUrl;
		this.mdFormat = mdFormat;
		this.set = set;
		this.fromDate = fromDate;
		this.untilDate = untilDate;
		started = false;
	}

	private void verifyStarted() {
		if (!started) {
			started = true;
			try {
				token = firstPage();
			} catch (final CollectorException e) {
				throw new RuntimeException(e);
			}
		}
	}

	@Override
	public boolean hasNext() {
		synchronized (queue) {
			verifyStarted();
			return !queue.isEmpty();
		}
	}

	@Override
	public String next() {
		synchronized (queue) {
			verifyStarted();
			final String res = queue.poll();
			while (queue.isEmpty() && (token != null) && !token.isEmpty()) {
				try {
					token = otherPages(token);
				} catch (final CollectorException e) {
					throw new RuntimeException(e);
				}
			}
			return res;
		}
	}

	@Override
	public void remove() {}

	private String firstPage() throws CollectorException {
		final StringBuilder url = new StringBuilder()
				.append(baseUrl)
				.append("?verb=ListRecords&metadataPrefix=")
				.append(URLEncoder.encode(mdFormat, StandardCharsets.UTF_8));

		if (StringUtils.isNotBlank(set)) {
			url.append("&set=").append(URLEncoder.encode(set, StandardCharsets.UTF_8));
		}
		if (fromDate != null) {
			url.append("&from=").append(URLEncoder.encode(fromDate.format(oaiDateFormatter), StandardCharsets.UTF_8));
		}
		if (untilDate != null) {
			url.append("&until=").append(URLEncoder.encode(untilDate.format(oaiDateFormatter), StandardCharsets.UTF_8));
		}
		return downloadPage(url.toString());
	}

	private String otherPages(final String resumptionToken) throws CollectorException {
		return downloadPage(baseUrl + "?verb=ListRecords&resumptionToken=" + URLEncoder.encode(resumptionToken, StandardCharsets.UTF_8));
	}

	private String extractResumptionToken(final String xml) {

		final String s = StringUtils.substringAfter(xml, "<resumptionToken");
		if (s == null) { return null; }

		final String result = StringUtils.substringBetween(s, ">", "</");
		if (result == null) { return null; }
		return result.trim();

	}

	private String downloadPage(final String url) throws CollectorException {
		final String xml = new RestTemplate().getForObject(url, String.class);
		Document doc;
		try {
			doc = reader.read(new StringReader(xml));
		} catch (final DocumentException e) {
			final String cleaned = XmlCleaner.cleanAllEntities(xml);
			try {
				doc = reader.read(new StringReader(cleaned));
			} catch (final DocumentException e1) {
				final String resumptionToken = extractResumptionToken(xml);
				if (resumptionToken == null) { throw new CollectorException("Error parsing cleaned document:" + cleaned, e1); }
				return resumptionToken;
			}
		}

		final Node errorNode = doc.selectSingleNode("/*[local-name()='OAI-PMH']/*[local-name()='error']");
		if (errorNode != null) {
			final String code = errorNode.valueOf("@code");
			if ("noRecordsMatch".equalsIgnoreCase(code.trim())) { return null; }
			throw new CollectorException(code + " - " + errorNode.getText());
		}

		for (final Object o : doc.selectNodes("//*[local-name()='ListRecords']/*[local-name()='record']")) {
			queue.add(((Node) o).asXML());
		}

		return doc.valueOf("//*[local-name()='resumptionToken']");

	}

}
